/*
 * Copyright (c) 2024. Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.antragraum;

import static java.util.Objects.*;

import java.time.ZonedDateTime;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import jakarta.annotation.PostConstruct;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.opensaml.saml.saml2.core.Response;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Service;

import de.ozgcloud.nachrichten.NachrichtenManagerConfiguration;
import de.ozgcloud.nachrichten.NachrichtenManagerProperties;
import de.ozgcloud.nachrichten.common.vorgang.Vorgang;
import de.ozgcloud.nachrichten.common.vorgang.VorgangService;
import de.ozgcloud.nachrichten.postfach.PersistPostfachNachrichtService;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtMapper;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Service
@RequiredArgsConstructor
@ConditionalOnProperty(AntragraumProperties.PROPERTY_ANTRAGSRAUM_URL)
public class AntragraumService {

	static final String BAYERN_ID_SERVICE_KONTO_TYPE = "BayernID";

	static final String USER_NOTIFICATION_TEMPLATE = """
			Guten Tag,

			bei der Bearbeitung Ihres Antrags hat sich eine Rückfrage ergeben. Bitte beantworten Sie diese im Antragsraum:

			%s

			Vielen Dank,
			Ihre digitale Verwaltung
			""";

	private final PersistPostfachNachrichtService postfachNachrichtService;
	private final Saml2Verifier verifier;
	private final Saml2Parser parser;
	private final Saml2Decrypter decrypter;

	private final AntragraumProperties properties;
	private final NachrichtenManagerProperties nachrichtenManagerProperties;
	private final RueckfrageMapper rueckfrageMapper;
	@Qualifier(NachrichtenManagerConfiguration.NACHRICHTEN_VORGANG_SERVICE)
	private final VorgangService vorgangService;
	private final PostfachNachrichtMapper nachrichtMapper;

	@PostConstruct
	void init() {
		if (isNull(nachrichtenManagerProperties.getAddress())) {
			throw new IllegalStateException("Address of Nachrichten-Manager is not set");
		}
	}

	@PostConstruct
	void logInitalization() {
		LOG.info("Starting Antragraum Service");
	}

	public String getAntragsraumUrl() {
		return properties.getUrl();
	}

	public String getUserNotificationText() {
		return USER_NOTIFICATION_TEMPLATE.formatted(getAntragsraumUrl());
	}

	public Stream<RueckfrageHead> findRueckfragen(String samlToken) {
		verifyToken(samlToken);

		var rueckfragen = postfachNachrichtService.findRueckfragen(BAYERN_ID_SERVICE_KONTO_TYPE, getPostfachId(samlToken))
				.map(rueckfrageMapper::fromPostfachNachricht)
				.map(rueckfrage -> enrichByAnswers(rueckfrage, samlToken)).toList();
		return enrichRueckfragen(rueckfragen, samlToken);
	}

	RueckfrageHead enrichByAnswers(RueckfrageHead rueckfrage, String samlToken) {
		var rueckfrageBuilder = rueckfrage.toBuilder();

		getFirstSentAt(findAnswers(samlToken, rueckfrage.getId()))
				.ifPresent(firstSentAt -> rueckfrageBuilder.answeredAt(firstSentAt).status(RueckfrageStatus.ANSWERED));

		return rueckfrageBuilder.build();
	}

	Optional<ZonedDateTime> getFirstSentAt(Stream<PostfachNachricht> answers) {
		return answers.sorted(Comparator.comparing(PostfachNachricht::getSentAt)).map(PostfachNachricht::getSentAt).findFirst();
	}

	Stream<RueckfrageHead> enrichRueckfragen(List<RueckfrageHead> rueckfragen, String samlToken) {
		if (rueckfragen.isEmpty()) {
			return Stream.empty();
		}
		var vorgang = vorgangService.getVorgang(rueckfragen.getFirst().getVorgangId());

		return rueckfragen.stream()
				.map(rueckfrage -> addVorgangData(rueckfrage, vorgang))
				.map(rueckfrage -> addAccessible(rueckfrage, samlToken));
	}

	RueckfrageHead addAccessible(RueckfrageHead rueckfrage, String samlToken) {
		return rueckfrage.toBuilder().accessible(isAccessible(samlToken, rueckfrage.getTrustLevel())).build();
	}

	RueckfrageHead addVorgangData(RueckfrageHead rueckfrage, Vorgang vorgang) {
		return rueckfrage.toBuilder()
				.vorgangName(vorgang.getName())
				.vorgangNummer(vorgang.getNummer())
				.trustLevel(vorgang.getTrustLevel())
				.build();
	}

	public String sendRueckfrageAnswer(String samlToken, String rueckfrageId, PostfachNachricht nachricht) {
		verifyToken(samlToken);

		return postfachNachrichtService.persistAnswer(rueckfrageId, nachricht);
	}

	public Stream<PostfachNachricht> findAnswers(String samlToken, String rueckfrageId) {
		verifyToken(samlToken);

		return postfachNachrichtService.findAnswers(BAYERN_ID_SERVICE_KONTO_TYPE, getPostfachId(samlToken), rueckfrageId);
	}

	String getPostfachId(String samlToken) {
		return decrypter.decryptPostfachId(parser.parse(samlToken));
	}

	public PostfachNachricht getRueckfrage(String samlToken, String id) {
		verifyToken(samlToken);

		return nachrichtMapper.fromMapToPostfachMail(postfachNachrichtService.getById(id));
	}

	void verifyToken(String token) {
		var errors = verifier.verify(token);
		if (CollectionUtils.isNotEmpty(errors)) {
			throw new SecurityException("SAML Token verification failed. Errors: %s".formatted(errors));
		}
	}

	public boolean isAccessible(String samlToken, String trustLevel) {
		return StringUtils.equals(getTrustLevel(samlToken), trustLevel);
	}

	String getTrustLevel(String samlToken) {
		return decrypter.decryptTrustLevel(parseSamlToken(samlToken));
	}

	Response parseSamlToken(String samlToken) {
		return parser.parse(samlToken);
	}
}