/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Stream;

import org.apache.logging.log4j.Logger;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mapstruct.factory.Mappers;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.boot.logging.LogLevel;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.nachrichten.antragraum.AntragraumService;
import de.ozgcloud.nachrichten.attributes.ClientAttributeService;
import de.ozgcloud.nachrichten.info.InfoManagerService;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht.Direction;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht.ReplyOption;
import de.ozgcloud.nachrichten.postfach.osi.MessageAttachmentService;
import de.ozgcloud.nachrichten.postfach.osi.MessageTestFactory;
import de.ozgcloud.nachrichten.postfach.osi.OsiPostfachServerProcessException;
import de.ozgcloud.nachrichten.postfach.osi.OsiPostfachServerProcessExceptionTestFactory;
import de.ozgcloud.vorgang.callcontext.CallContextUserTestFactory;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import nl.altindag.log.LogCaptor;

class PostfachServiceTest {

	@Spy
	@InjectMocks
	private PostfachService service;
	@Mock
	private PostfachRemoteService postfachRemoteService;

	@Mock
	private PersistPostfachNachrichtService persistingService;
	@Spy
	private PostfachNachrichtMapper mapper = Mappers.getMapper(PostfachNachrichtMapper.class);

	@Mock
	private ClientAttributeService clientAttributeService;
	@Mock
	private MessageAttachmentService messageAttachmentService;
	@Mock
	private ApplicationEventPublisher publisher;

	@Mock
	private CurrentUserService userService;
	@Spy
	private Optional<AntragraumService> antragsraumService = Optional.of(mock(AntragraumService.class));
	@Spy
	private Optional<InfoManagerService> infomanagerService = Optional.of(mock(InfoManagerService.class));

	static final String COMMAND_ID = UUID.randomUUID().toString();
	static final String USER_ID = UUID.randomUUID().toString();

	@Nested
	class TestSaveDraft {

		@Captor
		private ArgumentCaptor<Optional<String>> userIdCaptor;
		@Captor
		private ArgumentCaptor<PostfachNachricht> nachrichtCaptor;

		@BeforeEach
		void init() {
			doNothing().when(service).persistMail(any(), any());

			when(userService.getUser()).thenReturn(CallContextUserTestFactory.create());
		}

		@Test
		void shouldUseUserIdFromContext() {
			service.saveDraft(MessageTestFactory.VORGANG_ID, PostfachNachrichtTestFactory.create());

			verify(service).persistMail(userIdCaptor.capture(), any());
			assertThat(userIdCaptor.getValue()).contains(CallContextUserTestFactory.ID);
		}

		@Test
		void shouldPersistNachricht() {
			service.saveDraft(MessageTestFactory.VORGANG_ID, PostfachNachrichtTestFactory.create());

			verify(service).persistMail(any(), any());
		}

		@Test
		void shouldAddVorgangId() {
			service.saveDraft(MessageTestFactory.VORGANG_ID, PostfachNachrichtTestFactory.createBuilder().vorgangId(null).build());

			verify(service).persistMail(any(), nachrichtCaptor.capture());
			assertThat(nachrichtCaptor.getValue().getVorgangId()).isEqualTo(MessageTestFactory.VORGANG_ID);
		}
	}

	@Nested
	class TestSendMail {

		@Test
		void shouldCallDoSendMail() {
			when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);
			var mail = PostfachNachrichtTestFactory.create();

			service.sendMail(COMMAND_ID, USER_ID, mail);

			verify(service).doSendMail(mail);
		}

		@Test
		void shouldCallPersistSentMail() {
			when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);
			var mail = PostfachNachrichtTestFactory.create();
			doReturn(mail).when(service).addMailSentInformation(any(PostfachNachricht.class), any(SendPostfachNachrichtResponse.class));

			service.sendMail(COMMAND_ID, USER_ID, PostfachNachrichtTestFactory.create());

			verify(service).persistSentMail(USER_ID, mail);
		}

		@Test
		void shouldAddMailSentInformation() {
			var postfachMail = service.addMailSentInformation(PostfachNachrichtTestFactory.createBuilder().sentAt(null).sentSuccessful(null).build(),
					SendPostfachNachrichtResponseTestFactory.create());

			assertThat(postfachMail.getSentAt()).isNotNull();
			assertThat(postfachMail.getSentSuccessful()).isEqualTo(SendPostfachNachrichtResponseTestFactory.SENT_SUCCESSFUL);
			assertThat(postfachMail.getMessageCode()).isEqualTo(SendPostfachNachrichtResponseTestFactory.MESSAGE_CODE.getMessageCode());
		}

		@Test
		void shouldHandleSendMail() {
			when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);
			var mail = PostfachNachrichtTestFactory.create();
			service.sendMail(COMMAND_ID, USER_ID, mail);

			verify(service).handleSendMail(COMMAND_ID, mail);
		}

		@Nested
		class TestPersistSentMail {

			@Captor
			private ArgumentCaptor<PostfachNachricht> mailCaptor;

			@Test
			void shouldCallPersistingService() {
				var mail = PostfachNachrichtTestFactory.create();

				service.persistSentMail(USER_ID, mail);

				verify(persistingService).persistNachricht(any(), notNull());
			}

			@Test
			void shouldSetDirectionToIN() {
				service.persistSentMail(USER_ID, PostfachNachrichtTestFactory.createBuilder().direction(null).build());

				verify(persistingService).persistNachricht(any(), mailCaptor.capture());
				assertThat(mailCaptor.getValue().getDirection()).isEqualTo(Direction.OUT);
			}

			@Test
			void shouldSetCreatedAt() {
				service.persistSentMail(USER_ID, PostfachNachrichtTestFactory.createBuilder().createdAt(null).build());

				verify(persistingService).persistNachricht(any(), mailCaptor.capture());
				assertThat(mailCaptor.getValue().getCreatedAt()).isCloseTo(ZonedDateTime.now(), within(2, ChronoUnit.SECONDS));
			}

			@Test
			void shouldSetCreatedBy() {
				service.persistSentMail(USER_ID, PostfachNachrichtTestFactory.createBuilder().createdBy(null).build());

				verify(persistingService).persistNachricht(any(), mailCaptor.capture());
				assertThat(mailCaptor.getValue().getCreatedBy()).isEqualTo(USER_ID);
			}

			@Test
			void shouldSetUserId() {
				var mail = PostfachNachrichtTestFactory.create();

				service.persistSentMail(USER_ID, mail);

				verify(persistingService).persistNachricht(eq(Optional.of(USER_ID)), any());
			}

			@Test
			void shouldSetClientAttribute() {
				service.persistSentMail(USER_ID, PostfachNachrichtTestFactory.create());

				verify(clientAttributeService).setHasPostfachNachricht(MessageTestFactory.VORGANG_ID);
			}

			@Test
			void shouldNOTSetNewNachricht() {
				service.persistSentMail(USER_ID, PostfachNachrichtTestFactory.create());

				verify(clientAttributeService, never()).setHasNewPostfachNachricht(any());
			}
		}

		@Nested
		class TestProcessForAntragsraum {

			@Mock
			private PostfachNachricht modifiedMail;

			@Test
			void shouldReturnEmpty() {
				doReturn(false).when(service).isPostfachWithAntragsraum(any());

				var result = service.processForAntragsraum(PostfachNachrichtTestFactory.create());

				assertThat(result).isEmpty();
			}

			@Test
			void shouldCallNotifyAntragsraum() {
				doReturn(true).when(service).isPostfachWithAntragsraum(any());
				var mail = PostfachNachrichtTestFactory.create();

				service.processForAntragsraum(mail);

				verify(infomanagerService.get()).notifyInfoManager(mail);
			}

			@Test
			void shouldReturnAdjustedMail() {
				doReturn(true).when(service).isPostfachWithAntragsraum(any());
				doReturn(Optional.of(modifiedMail)).when(service).adjustMail(any());

				var result = service.processForAntragsraum(PostfachNachrichtTestFactory.create());

				assertThat(result).contains(modifiedMail);
			}

			@DisplayName("Is notify antragsraum")
			@Nested
			class TestIsNotifyAntragsraum {

				@DisplayName("should return true if")
				@ParameterizedTest(name = "replyOption is {0}")
				@EnumSource(value = ReplyOption.class, names = { "FORBIDDEN" }, mode = EnumSource.Mode.EXCLUDE)
				void shouldConfirm(ReplyOption replyOption) {
					doReturn(true).when(service).isPostfachWithAntragsraum(any());

					var result = service.isNotifyAntragsraum(replyOption);

					assertThat(result).isTrue();
				}

				@DisplayName("should return false if reply is forbidden")
				@Test
				void shouldDeclineIfForbidden() {
					var result = service.isNotifyAntragsraum(ReplyOption.FORBIDDEN);

					assertThat(result).isFalse();
				}

				@DisplayName("should return false if antragsraum is not configured")
				@Test
				void shouldDeclineIfNotConfigured() {
					doReturn(false).when(service).isPostfachWithAntragsraum(any());

					var result = service.isNotifyAntragsraum(ReplyOption.POSSIBLE);

					assertThat(result).isFalse();
				}
			}

			@Nested
			class TestAdjustMail {

				@Mock
				private PostfachNachricht mail;

				@Nested
				class TestReplaceBody {

					@Test
					void shouldReturnEmptyIfNoAntragsraum() {
						ReflectionTestUtils.setField(service, "antragraumService", Optional.empty());

						var result = adjustMail();

						assertThat(result).isEmpty();
					}

					@Test
					void shouldAdjustMailBody() {
						when(mail.toBuilder()).thenReturn(PostfachNachrichtTestFactory.createBuilder());
						var expectedMailBody = "Antragsraum Text";
						when(antragsraumService.get().getUserNotificationText()).thenReturn(expectedMailBody);

						var result = adjustMail();

						assertThat(result).isPresent().get().extracting(PostfachNachricht::getMailBody).isEqualTo(expectedMailBody);
					}
				}

				private Optional<PostfachNachricht> adjustMail() {
					return service.adjustMail(mail);
				}
			}

		}
	}

	@Nested
	class TestFindById {
		@Test
		void shouldCallPersistingService() {
			service.findById(PostfachNachrichtTestFactory.ID);

			verify(persistingService).findById(PostfachNachrichtTestFactory.ID);
		}
	}

	@Nested
	class TestFindByVorgang {

		private static final String VORGANG_ID = UUID.randomUUID().toString();

		@Test
		void shouldCallPersistingService() {
			service.findByVorgang(VORGANG_ID);

			verify(persistingService).findByVorgangAsMap(VORGANG_ID);
		}
	}

	@Nested
	class TestFetchAndPersistReplies {

		private static final PostfachNachricht nachricht = PostfachNachrichtTestFactory.create();

		@BeforeEach
		void initTest() {
			lenient().when(postfachRemoteService.getAllMessages()).thenReturn(Stream.of(nachricht));
		}

		@Test
		void shouldCallGetAllMessages() {
			service.fetchAndPersistReplies();

			verify(postfachRemoteService).getAllMessages();
		}

		@Test
		void shouldCallPersistingService() {
			service.fetchAndPersistReplies();

			verify(persistingService).persistNachricht(eq(Optional.empty()), any());
		}

		@Test
		void shouldCallDelete() {
			service.fetchAndPersistReplies();

			verify(postfachRemoteService).deleteMessage(MessageTestFactory.MESSAGE_ID);
		}

		@Test
		void shouldThrowNotConfiguredException() {
			doReturn(false).when(service).isPostfachConfigured();

			assertThatExceptionOfType(NotConfiguredException.class).isThrownBy(() -> service.fetchAndPersistReplies());
			verify(postfachRemoteService, never()).getAllMessages();
		}
	}

	@Nested
	class TestPeristMail {
		@Test
		void shouldPersistMail() {
			PostfachNachricht mail = PostfachNachrichtTestFactory.create();

			service.persistMail(Optional.of(USER_ID), mail);

			verify(persistingService).persistNachricht(eq(Optional.of(USER_ID)), same(mail));
		}

		@Test
		void shouldSetHasPostfachNachricht() {
			var mail = PostfachNachrichtTestFactory.create();

			service.persistMail(Optional.of(USER_ID), mail);

			verify(clientAttributeService).setHasPostfachNachricht(MessageTestFactory.VORGANG_ID);
		}

		@Test
		@DisplayName("should NOT set hasPostfachNachricht if Nachricht is sent from system")
		void shouldNOTSetHasNachrichtForSystemNachricht() {
			var mail = PostfachNachrichtTestFactory.createBuilder().createdBy("system-123").build();

			service.persistMail(Optional.of(USER_ID), mail);

			verify(clientAttributeService, never()).setHasPostfachNachricht(MessageTestFactory.VORGANG_ID);
		}
	}

	@Nested
	class TestResendPostfachMail {

		static final String COMMAND_ID = UUID.randomUUID().toString();
		static final PostfachNachricht mail = PostfachNachrichtTestFactory.create();

		@BeforeEach
		void mockService() {
			when(persistingService.getById(anyString())).thenReturn(PostfachNachrichtTestFactory.asMap());
			doReturn(mail).when(mapper).fromMapToPostfachMail(anyMap());
			when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);
		}

		@Test
		void shouldCallPersistingServiceGetById() {
			service.resendMail(COMMAND_ID, PostfachNachrichtTestFactory.ID);

			verify(persistingService).getById(PostfachNachrichtTestFactory.ID);
		}

		@Test
		void shouldMapFromMapToPostfachMail() {
			service.resendMail(COMMAND_ID, PostfachNachrichtTestFactory.ID);

			verify(mapper).fromMapToPostfachMail(PostfachNachrichtTestFactory.asMap());
		}

		@Test
		void shouldSendMail() {
			service.resendMail(COMMAND_ID, PostfachNachrichtTestFactory.ID);

			verify(postfachRemoteService).sendMessage(any(PostfachNachricht.class));
		}

		@Test
		void shouldPatchMail() {
			service.resendMail(COMMAND_ID, PostfachNachrichtTestFactory.ID);

			verify(persistingService).patch(anyString(), anyMap());
		}

		@Test
		void shouldHandleSentMail() {
			service.resendMail(COMMAND_ID, PostfachNachrichtTestFactory.ID);

			verify(service).handleSendMail(COMMAND_ID, mail);
		}
	}

	@Nested
	class TestCreateResendPatchMap {

		@Test
		void shouldContainsSentAt() {
			var map = service.createResendPatchMap(SendPostfachNachrichtResponseTestFactory.create());

			assertThat(map).containsKey(PostfachNachricht.FIELD_SENT_AT);
			assertThat(ZonedDateTime.parse(map.get(PostfachNachricht.FIELD_SENT_AT).toString())).isCloseTo(ZonedDateTime.now(),
					within(2, ChronoUnit.SECONDS));
		}

		@Test
		void shouldContainsSentSuccessful() {
			var map = service.createResendPatchMap(SendPostfachNachrichtResponseTestFactory.create());

			assertThat(map).containsEntry(PostfachNachricht.FIELD_SENT_SUCCESSFUL, true);
		}

		@Test
		void shouldContainsMessageCode() {
			var map = service.createResendPatchMap(SendPostfachNachrichtResponseTestFactory.create());

			assertThat(map).containsEntry(PostfachNachricht.FIELD_MESSAGE_CODE,
					SendPostfachNachrichtResponseTestFactory.MESSAGE_CODE.getMessageCode());
		}
	}

	@DisplayName("Handle send mail")
	@Nested
	class TestHandleSendMail {

		final String COMMAND_ID = UUID.randomUUID().toString();
		final PostfachNachricht mail = PostfachNachrichtTestFactory.create();

		@Test
		void shouldDoSendMail() {
			when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);

			service.handleSendMail(COMMAND_ID, mail);

			verify(service).doSendMail(mail);
		}

		@Nested
		class TestSentSuccess {

			@Captor
			private ArgumentCaptor<PostfachMailSentEvent> eventCaptor;

			@Mock
			private PostfachNachricht modifiedMail;

			@Test
			void shouldDoSendMail() {
				when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);

				service.handleSendMail(COMMAND_ID, mail);

				verify(service).doSendMail(mail);
			}

			@Test
			void shouldPublishEvent() {
				when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);

				service.handleSendMail(COMMAND_ID, mail);

				verify(publisher).publishEvent(eventCaptor.capture());
				assertThat(eventCaptor.getValue().getSource()).isEqualTo(COMMAND_ID);
			}

			@Test
			void shouldReturnResponse() {
				when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);

				var response = service.handleSendMail(COMMAND_ID, mail);

				assertThat(response.isSentSuccessful()).isTrue();
				assertThat(response.getMessageCode()).isEqualTo(PostfachMessageCode.SEND_SUCCESSFUL_MESSAGE_CODE);
			}

			@Test
			void shouldSendProcessedPostfachNachricht() {
				doReturn(Optional.of(modifiedMail)).when(service).processForAntragsraum(any());

				service.handleSendMail(COMMAND_ID, mail);

				verify(service).doSendMail(modifiedMail);
			}

		}

		@DisplayName("receive")
		@Nested
		class TestSentFailed {

			@Captor
			private ArgumentCaptor<PostfachMailSentFailedEvent> eventFailedCaptor;

			@DisplayName("OsiPostfachException")
			@Nested
			class TestOnOsiPostfachException {

				private final static String MESSAGE = "Osi Postfach throws an exception.";

				@BeforeEach
				void mockService() {
					when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);
					doThrow(new PostfachException(MESSAGE, PostfachMessageCode.SEND_SUCCESSFUL_MESSAGE_CODE)).when(postfachRemoteService)
							.sendMessage(any());
				}

				@Test
				void shouldPublishEvent() {
					service.handleSendMail(COMMAND_ID, mail);

					verify(publisher).publishEvent(eventFailedCaptor.capture());
					assertThat(eventFailedCaptor.getValue().getSource()).isEqualTo(COMMAND_ID);
					assertThat(eventFailedCaptor.getValue().getErrorMessage()).startsWith(MESSAGE);
				}

				@Test
				void shouldLogError() {
					var logCaptor = LogCaptor.forClass(PostfachService.class);

					service.handleSendMail(COMMAND_ID, mail);

					verify(service).proceedWithErrorException(eq(COMMAND_ID), any(PostfachException.class));
					assertThat(logCaptor.getLogEvents().get(0).getLevel()).isEqualTo(LogLevel.ERROR.name());
					assertThat(logCaptor.getLogEvents().get(0).getMessage()).startsWith(MESSAGE);
				}

				@Test
				void shouldReturnResponse() {
					var response = service.handleSendMail(COMMAND_ID, mail);

					assertThat(response.isSentSuccessful()).isFalse();
					assertThat(response.getMessageCode()).isEqualTo(PostfachMessageCode.SEND_SUCCESSFUL_MESSAGE_CODE);
				}
			}

			@DisplayName("OsiPostfachServerProcessException")
			@Nested
			class TestOsiPostfachServerProcessException {

				private final static String MESSAGE = "Postfach server returned false";

				@Mock
				private Logger logger;

				@BeforeEach
				void mockService() {
					doThrow(OsiPostfachServerProcessExceptionTestFactory.create()).when(postfachRemoteService).sendMessage(any());
					when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);
				}

				@Test
				void shouldPublishEvent() {
					service.handleSendMail(COMMAND_ID, mail);

					verify(publisher).publishEvent(eventFailedCaptor.capture());
					assertThat(eventFailedCaptor.getValue().getSource()).isEqualTo(COMMAND_ID);
					assertThat(eventFailedCaptor.getValue().getErrorMessage()).startsWith(MESSAGE);
				}

				@Test
				void shouldLogWarning() {
					var logCaptor = LogCaptor.forClass(PostfachService.class);

					service.handleSendMail(COMMAND_ID, mail);

					verify(service).proceedwithWarnException(eq(COMMAND_ID), any(OsiPostfachServerProcessException.class));
					assertThat(logCaptor.getLogEvents().get(0).getLevel()).isEqualTo(LogLevel.WARN.name());
					assertThat(logCaptor.getLogEvents().get(0).getMessage()).startsWith(MESSAGE);
				}

				@Test
				void shouldReturnResponse() {
					var response = service.handleSendMail(COMMAND_ID, mail);

					assertThat(response.isSentSuccessful()).isFalse();
					assertThat(response.getMessageCode()).isEqualTo(PostfachMessageCode.PROCESS_FAILED_MESSAGE_CODE);
				}
			}
		}

	}

	@Nested
	class TestDoSendMail {
		@Test
		void shouldCallPostfachRemoteService() {
			service.doSendMail(PostfachNachrichtTestFactory.create());

			verify(postfachRemoteService).sendMessage(any());
		}

		@Nested
		class WithoutRemoteService {
			@InjectMocks
			private PostfachService service;

			@Test
			void shouldThrowNotConfiguredException() {
				var nachricht = PostfachNachrichtTestFactory.create();

				assertThatExceptionOfType(NotConfiguredException.class).isThrownBy(() -> service.doSendMail(nachricht));

				verify(postfachRemoteService, never()).sendMessage(any());
			}
		}

	}

	@Nested
	class TestIsPostfachConfigured {

		@Test
		void shouldReturnTrue() {
			assertThat(service.isPostfachConfigured()).isTrue();
		}
	}

	@Nested
	class TestGetPostfachs {

		private Postfach postfach = PostfachTestFactory.create();

		@Test
		void shouldCallBuildPostfach() {
			when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);

			service.getPostfachs();

			verify(service).buildPostfach(postfachRemoteService);
		}

		@Test
		void shouldReturnPostfachs() {
			doReturn(postfach).when(service).buildPostfach(any());

			var result = service.getPostfachs();

			assertThat(result).containsOnly(postfach);
		}

		@Test
		void shouldReturnEmptyStream() {
			doReturn(false).when(service).isPostfachConfigured();

			var result = service.getPostfachs();

			assertThat(result).isEmpty();
		}

		@Nested
		class TestBuildPostfach {

			@Test
			void shouldSetPostfachType() {
				when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);

				var result = service.buildPostfach(postfachRemoteService);

				assertThat(result.getType()).isEqualTo(PostfachTestFactory.POSTFACH_TYPE);
			}

			@Test
			void shouldSetIsReplyAllowed() {
				doReturn(true).when(service).isReplyAllowed(any());

				var result = service.buildPostfach(postfachRemoteService);

				assertThat(result.isReplyAllowed()).isTrue();
			}
		}

		@Nested
		class TestIsReplyAllowed {

			@Test
			void shouldAllowWithAntragsraum() {
				doReturn(true).when(service).isPostfachWithAntragsraum(postfachRemoteService);

				var result = service.isReplyAllowed(postfachRemoteService);

				assertThat(result).isTrue();
			}

			@Test
			void shouldAllowWithPostfach() {
				doReturn(false).when(service).isPostfachWithAntragsraum(postfachRemoteService);
				when(postfachRemoteService.isReplyAllowed()).thenReturn(true);

				var result = service.isReplyAllowed(postfachRemoteService);

				assertThat(result).isTrue();
			}

			@Test
			void shouldNotAllow() {
				doReturn(false).when(service).isPostfachWithAntragsraum(postfachRemoteService);

				var result = service.isReplyAllowed(postfachRemoteService);

				assertThat(result).isFalse();
			}
		}
	}
}